@@ -107,6 +107,7 @@ use /u/ben/work/scuzzy-ssleay8/ssl /u/ben/work/scuzzy-ssleay8/crypto/bio
 #if CACHE_SESSIONS
 # include "gcache.h"
 static const char *s_szCacheServerPort;
+static const char *s_szCacheServerDir;
 static char *s_szCacheServerPath;
 
 #define DEBUG_SESSIONS		0
@@ -195,30 +196,38 @@ static const char six2pr[64+1]=
   "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
 
 static void uuencoden(unsigned char *szTo,const unsigned char *szFrom,
-                     int nLength)
+                     int nLength,BOOL bPad)
     {
     const unsigned char *s;
+    int nPad=0;
 
     for(s=(const unsigned char *)szFrom ; nLength > 0 ; s+=3)
 	{
 	*szTo++=six2pr[s[0] >> 2];
 	*szTo++=six2pr[(s[0] << 4 | s[1] >> 4)&0x3f];
 	if(!--nLength)
+	    {
+	    nPad=2;
 	    break;
+	    }
 	*szTo++=six2pr[(s[1] << 2 | s[2] >> 6)&0x3f];
 	if(!--nLength)
+	    {
+	    nPad=1;
 	    break;
+	    }
 	*szTo++=six2pr[s[2]&0x3f];
-	if(!--nLength)
-	    break;
+	--nLength;
 	}
+    while(bPad && nPad--)
+	*szTo++='=';
     *szTo++='\0';
     }
 
-static void uuencode(char *szTo,const char *szFrom)
+static void uuencode(char *szTo,const char *szFrom,BOOL bPad)
     {
     uuencoden((unsigned char *)szTo,(const unsigned char *)szFrom,
-	      strlen(szFrom));
+	      strlen(szFrom),bPad);
     }
 
 static int SSLCheckCipher(request_rec *r)
@@ -306,8 +315,8 @@ static void AddCertToEnv(pool *p,table *e,char *szEnv,X509 *xs)
     n=i2d_X509(xs,NULL);
     d=s=ap_palloc(p,n);
     i2d_X509(xs,&d);
-    d=ap_palloc(p,(n*4)/3+2);
-    uuencoden(d,s,n);
+    d=ap_palloc(p,(n*4)/3+2+2);
+    uuencoden(d,s,n,TRUE);
     ap_table_setn(e,szEnv,(char *)d);
     }
 #endif
@@ -412,7 +421,7 @@ static int SSLFixups(request_rec *r)
 	ExpandCert(r->pool,e,"SSL_CLIENT_",
 		   r->connection->client->szClientX509);
 	xs=SSL_get_peer_certificate(r->connection->client->ssl);
-	ExpandCert(r->pool,e,"SSL_CLIENT_I",
+	ExpandCert(r->pool,e,"SSL_CLIENT_I_",
 		   X509_NAME_oneline(X509_get_issuer_name(xs),NULL,0));
 #if APACHE_SSL_EXPORT_CERTS
 	if(pDirConfig->bExportCerts)
@@ -423,7 +432,7 @@ static int SSLFixups(request_rec *r)
     xs=SSL_get_certificate(r->connection->client->ssl);
     ExpandCert(r->pool,e,"SSL_SERVER_",
 	       X509_NAME_oneline(X509_get_subject_name(xs),NULL,0));
-    ExpandCert(r->pool,e,"SSL_SERVER_I",
+    ExpandCert(r->pool,e,"SSL_SERVER_I_",
 	       X509_NAME_oneline(X509_get_issuer_name(xs),NULL,0));
 
 #if APACHE_SSL_EXPORT_CERTS
@@ -647,7 +656,10 @@ static int ApacheSSLVerifyCallback(int ok,X509 *xs,X509 *xi,int depth,
 	}
 #endif
 
-    if(error == X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY || error == X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE)
+    if(error == X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY
+       || error == X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE
+       || error == X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT
+       || error == X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN)
 	{
 	char *s;
 
@@ -665,7 +677,7 @@ static int ApacheSSLVerifyCallback(int ok,X509 *xs,X509 *xi,int depth,
 	if(s == NULL)
 	    {
 	    SSLLogError(pServer,APLOG_ERR,"verify error");
-	    return(0);
+	    return FALSE;
 	    }
 	ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,"issuer=%s",
 		     s);
@@ -687,7 +699,7 @@ static int ApacheSSLVerifyCallback(int ok,X509 *xs,X509 *xi,int depth,
 	}
     ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,
 		 "verify return:%d",ok);
-    return(ok);
+    return ok;
     }
 
 static int VerifyFlags(SSLConfigRec *pConfig)
@@ -1079,6 +1091,13 @@ static int CacheLauncher(void *cmd,child_info *pInfo)
     dup2(STDERR_FILENO,STDOUT_FILENO);
     ap_snprintf(a1,sizeof a1,"%d",ap_user_id);
     ap_cleanup_for_exec();
+    if(s_szCacheServerDir)
+	{
+	if(chdir(s_szCacheServerDir) < 0)
+	    perror(s_szCacheServerDir);
+	}
+    ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,s_pServer,
+		 "gcache pid is %d",getpid());
     execl(cmd,"gcache",a1,s_szCacheServerPort,NULL);
     perror(cmd);
     assert(0);
@@ -1221,7 +1240,7 @@ static int FakeBasicAuth(request_rec *r)
 */
     ap_snprintf(b1,sizeof b1,"%s:password",
 		r->connection->client->szClientX509);
-    uuencode(b2,b1);
+    uuencode(b2,b1,FALSE);
     t=ap_psprintf(r->pool,"Basic %s",b2);
     ap_table_setn(r->headers_in,"Authorization",t);
     return DECLINED;
@@ -1418,6 +1437,14 @@ static const char *SSLCacheServerPath(cmd_parms *cmd,char *struct_ptr,
     {
     s_szCacheServerPath=ap_server_root_relative(cmd->pool,arg);
 
+    return NULL;
+    }
+
+static const char *SSLCacheServerDir(cmd_parms *cmd,char *struct_ptr,
+				     char *arg)
+    {
+    s_szCacheServerDir=ap_server_root_relative(cmd->pool,arg);
+
     return NULL;
     }
 #endif
@@ -1517,6 +1544,8 @@ command_rec ssl_cmds[]=
       "Lifetime in seconds of session cache objects" },
     { "SSLCacheServerPath",SSLCacheServerPath,NULL,RSRC_CONF,TAKE1,
       "Path of the global cache server" },
+    { "SSLCacheServerRunDir",SSLCacheServerDir,NULL,RSRC_CONF,TAKE1,
+      "Path for the server to run in (in case of coredumps)" },
 #endif
     { "SSLLogFile",SSLLogFile,NULL,RSRC_CONF,TAKE1,
       "Defunct SSL log file - DO NOT USE" },
